import base64
import pickle
from typing import Union

from django.db.models import QuerySet

from django_glue.access.access import GlueAccess
from django_glue.entities.base_entity import GlueEntity
from django_glue.entities.model_object.fields.factories import model_object_fields_from_model
from django_glue.entities.query_set.session_data import GlueQuerySetSessionData
from django_glue.handler.enums import GlueConnection


class GlueQuerySet(GlueEntity):
    def __init__(
            self,
            unique_name: str,
            query_set: QuerySet,
            access: Union[GlueAccess, str] = GlueAccess.VIEW,
            included_fields: tuple = ('__all__',),
            excluded_fields: tuple = ('__none__',),
            included_methods: tuple = ('__none__',),
    ):
        super().__init__(unique_name, GlueConnection.QUERY_SET, access)

        self.query_set = query_set
        self.model = query_set.model

        self.included_fields = included_fields
        self.excluded_fields = excluded_fields
        self.included_methods = included_methods

    def encode_query_set(self):
        return base64.b64encode(pickle.dumps(self.query_set.query)).decode()

    def to_session_data(self) -> GlueQuerySetSessionData:
        return GlueQuerySetSessionData(
            unique_name=self.unique_name,
            query_set_str=self.encode_query_set(),
            connection=self.connection,
            fields=model_object_fields_from_model(self.model, self.included_fields, self.excluded_fields).to_dict(),
            access=self.access,
            included_fields=self.included_fields,
            excluded_fields=self.excluded_fields,
            included_methods=self.included_methods,
        )



from django_glue.entities.query_set.actions import GlueQuerySetAction
from django_glue.entities.query_set import handlers


GLUE_QUERY_SET_HANDLER_MAP = {
    GlueQuerySetAction.ALL: handlers.AllGlueQuerySetHandler,
    GlueQuerySetAction.DELETE: handlers.DeleteGlueQuerySetHandler,
    GlueQuerySetAction.FILTER: handlers.FilterGlueQuerySetHandler,
    GlueQuerySetAction.GET: handlers.GetGlueQuerySetHandler,
    GlueQuerySetAction.NULL_OBJECT: handlers.NullObjectGlueQuerySetHandler,
    GlueQuerySetAction.METHOD: handlers.MethodGlueQuerySetHandler,
    GlueQuerySetAction.UPDATE: handlers.UpdateGlueQuerySetHandler,
    GlueQuerySetAction.TO_CHOICES: handlers.ToChoicesGlueQuerySetHandler,
}



from django_glue.form.field.attrs import factories
from django_glue.form.field.enums import FieldType


FIELD_TYPE_TO_GLUE_ATTR_FACTORY = {
    FieldType.AUTO_FIELD: factories.GlueIntegerAttrFactory,
    FieldType.BIG_AUTO_FIELD: factories.GlueIntegerAttrFactory,
    FieldType.SMALL_AUTO_FIELD: factories.GlueIntegerAttrFactory,
    FieldType.BOOLEAN: factories.GlueBooleanAttrFactory,
    FieldType.CHAR: factories.GlueCharAttrFactory,
    FieldType.COMMA_SEPARATED_INTEGER: factories.GlueIntegerAttrFactory,  # Now depreciated in django.
    FieldType.DATE: factories.GlueDateAttrFactory,
    FieldType.DATETIME: factories.GlueDateAttrFactory,
    FieldType.DECIMAL: factories.GlueDecimalAttrFactory,
    FieldType.DURATION: factories.GlueCharAttrFactory,
    FieldType.EMAIL: factories.GlueCharAttrFactory,
    FieldType.FLOAT: factories.GlueDecimalAttrFactory,
    FieldType.INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.BIG_INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.SMALL_INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.GENERIC_IP_ADDRESS: factories.GlueCharAttrFactory,
    FieldType.IP_ADDRESS: factories.GlueCharAttrFactory,
    FieldType.FOREIGN_KEY: factories.GlueIntegerAttrFactory,
    FieldType.ONE_TO_ONE: factories.GlueIntegerAttrFactory,
    FieldType.MANY_TO_MANY: factories.GlueCharAttrFactory,
    FieldType.JSON: factories.GlueTextAreaAttrFactory,
    FieldType.NULL_BOOLEAN: factories.GlueBooleanAttrFactory,  # Now depreciated in django
    FieldType.POSITIVE_BIG_INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.POSITIVE_INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.POSITIVE_SMALL_INTEGER: factories.GlueIntegerAttrFactory,
    FieldType.SLUG: factories.GlueCharAttrFactory,
    FieldType.TEXT: factories.GlueTextAreaAttrFactory,
    FieldType.TIME: factories.GlueDateAttrFactory,
    FieldType.URL: factories.GlueCharAttrFactory,
    FieldType.UUID: factories.GlueCharAttrFactory
}



from django.urls import path

from django_glue import views

app_name = 'django_glue'

urlpatterns = [
    # These url path names are used in our middleware to avoid cleaning session data.
    path("", views.glue_data_ajax_handler_view, name="django_glue_data_handler"),
    path("keep_live/", views.glue_keep_live_handler_view, name="django_glue_keep_live_handler"),
    path("session/", views.glue_session_data_view, name="django_glue_session_data"),
]



from dataclasses import dataclass
from typing import Any

from django_glue.entities.model_object.fields.entities import GlueModelFields
from django_glue.response.data import GlueJsonData


@dataclass
class GlueModelObjectJsonData(GlueJsonData):  # This is a little duplicated but allows us to send more response data.
    fields: GlueModelFields

    def to_dict(self):
        return self.fields.to_dict()


@dataclass
class MethodGlueModelObjectJsonData(GlueJsonData):
    method_return: Any

    def to_dict(self):
        return {'method_return': self.method_return}



from django_glue.access.access import GlueAccess
from django_glue.entities.base_entity import GlueEntity
from django_glue.entities.function.response_data import GlueFunctionJsonData
from django_glue.entities.function.session_data import FunctionSessionData
from django_glue.handler.enums import GlueConnection
from django_glue.utils import check_valid_method_kwargs, type_set_method_kwargs


class GlueFunction(GlueEntity):
    def __init__(
            self,
            unique_name: str,
            function_path: str,
    ):
        super().__init__(unique_name, GlueConnection.FUNCTION, GlueAccess.VIEW)
        self.function_path = function_path

        self.module_name = '.'.join(function_path.split('.')[:-1])
        self.function_name = function_path.split('.')[-1]

    def call(self, function_kwargs):
        module = __import__(self.module_name, fromlist=[self.function_name])

        if hasattr(module, self.function_name):
            function = getattr(module, self.function_name)

            if check_valid_method_kwargs(function, function_kwargs):
                type_set_kwargs = type_set_method_kwargs(function, function_kwargs)

                return function(**type_set_kwargs)

        return None

    def to_session_data(self) -> FunctionSessionData:
        return FunctionSessionData(
            unique_name=self.unique_name,
            connection=self.connection,
            access=self.access,
            function_path=self.function_path
        )

    def to_response_data(self, function_return) -> GlueFunctionJsonData:
        return GlueFunctionJsonData(function_return)



from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any

from django.db.models import Field

from django_glue.form.field.attrs.entities import GlueFieldAttrs, GlueFieldAttr


class GlueAttrFactory(ABC):
    def __init__(self, model_field: Field):
        self.model_field = model_field
        self.glue_field_attrs = GlueFieldAttrs()

    def add_attr(
            self,
            name: str,
            value: Any,
    ) -> None:
        attr = GlueFieldAttr(name=name, value=value)
        self.glue_field_attrs += attr

    @abstractmethod
    def add_field_attrs(self):
        pass

    def add_base_attrs(self):
        self.add_attr('name', self.model_field.name)
        self.add_attr('id', f'id_{self.model_field.name}')

        if not self.model_field.blank:
            self.add_attr('required', True)

        if self.model_field.hidden:
            self.add_attr('hidden', True)

        if self.model_field.max_length:
            self.add_attr('maxlength', self.model_field.max_length)

    def factory_method(self) -> GlueFieldAttrs:
        self.glue_field_attrs = GlueFieldAttrs()
        self.add_base_attrs()
        self.add_field_attrs()
        return self.glue_field_attrs


class GlueBooleanAttrFactory(GlueAttrFactory):
    def add_field_attrs(self):
        pass


class GlueCharAttrFactory(GlueAttrFactory):
    def add_field_attrs(self):
        pass


class GlueDateAttrFactory(GlueAttrFactory):
    def add_field_attrs(self):
        self.add_attr('max', '')
        self.add_attr('min', '')


class GlueTextAreaAttrFactory(GlueAttrFactory):
    def add_field_attrs(self):
        self.add_attr('cols', 20)
        self.add_attr('rows', 3)

        if self.model_field.max_length:
            self.add_attr('maxlength', self.model_field.max_length)


class GlueIntegerAttrFactory(GlueTextAreaAttrFactory):
    def add_field_attrs(self):
        self.max_min_validation_attr()
        self.step_attr()

    def max_min_validation_attr(self):
        """
            The range is between Number.MIN_SAFE_INTEGER and Number.MAX_SAFE_INTEGER
            represents Javascript integer limits.
        """
        MIN_SAFE_INTEGER = -9007199254740991
        MAX_SAFE_INTEGER = 9007199254740991
        min_max_validators = {'min_value', 'max_value'}

        for validator in self.model_field.validators:
            if hasattr(validator, 'code') and validator.code in min_max_validators:
                attr_name = validator.code.split('_')[0]

                limit_value = (
                    validator.limit_value
                    if MIN_SAFE_INTEGER <= validator.limit_value <= MAX_SAFE_INTEGER
                    else None
                )

                self.add_attr(attr_name, limit_value)

    def step_attr(self):
        self.add_attr('step', 1)


class GlueDecimalAttrFactory(GlueIntegerAttrFactory):
    def add_field_attrs(self):
        super().add_field_attrs()

    def step_attr(self):
        validator = next((v for v in self.model_field.validators if hasattr(v, 'decimal_places')), None)

        if validator:
            step_value = Decimal('1') / (10 ** validator.decimal_places)
            self.add_attr('step', float(step_value))
        else:
            self.add_attr('step', 0.01)



from django_glue.access.decorators import check_access
from django_glue.entities.model_object.actions import GlueModelObjectAction
from django_glue.entities.post_data import UpdatePostData, MethodPostData
from django_glue.entities.model_object.factories import glue_model_object_from_glue_session
from django_glue.entities.model_object.response_data import MethodGlueModelObjectJsonData, GlueModelObjectJsonData
from django_glue.entities.model_object.session_data import GlueModelObjectSessionData
from django_glue.handler.handlers import GlueRequestHandler
from django_glue.response.data import GlueJsonResponseData
from django_glue.response.responses import generate_json_200_response_data


class GetGlueModelObjectHandler(GlueRequestHandler):
    action = GlueModelObjectAction.GET
    _session_data_class = GlueModelObjectSessionData

    @check_access
    def process_response_data(self) -> GlueJsonResponseData:
        glue_model_object = glue_model_object_from_glue_session(self.session_data)
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=GlueModelObjectJsonData(glue_model_object.fields)
        )


class UpdateGlueModelObjectHandler(GlueRequestHandler):
    action = GlueModelObjectAction.UPDATE
    _session_data_class = GlueModelObjectSessionData
    _post_data_class = UpdatePostData

    @check_access
    def process_response_data(self) -> GlueJsonResponseData:
        glue_model_object = glue_model_object_from_glue_session(self.session_data)
        glue_model_object.update(self.post_data.fields)
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully updated model object!',
            data=GlueModelObjectJsonData(glue_model_object.fields)
        )


class DeleteGlueModelObjectHandler(GlueRequestHandler):
    action = GlueModelObjectAction.DELETE
    _session_data_class = GlueModelObjectSessionData

    @check_access
    def process_response_data(self) -> GlueJsonResponseData:
        glue_model_object = glue_model_object_from_glue_session(self.session_data)
        glue_model_object.model_object.delete()
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully deleted model object!',
        )


class MethodGlueModelObjectHandler(GlueRequestHandler):
    action = GlueModelObjectAction.METHOD
    _session_data_class = GlueModelObjectSessionData
    _post_data_class = MethodPostData

    @check_access
    def process_response_data(self) -> GlueJsonResponseData:
        glue_model_object = glue_model_object_from_glue_session(self.session_data)
        method_return = glue_model_object.call_method(self.post_data.method, self.post_data.kwargs)
        return generate_json_200_response_data(
            'THE METHOD ACTION',
            'this is a response from an model object method action!',
            data=MethodGlueModelObjectJsonData(method_return)
        )



from django_glue.handler.enums import GlueConnection
from django_glue.entities.function.maps import GLUE_FUNCTION_HANDLER_MAP
from django_glue.entities.model_object.maps import GLUE_MODEL_OBJECT_HANDLER_MAP
from django_glue.entities.query_set.maps import GLUE_QUERY_SET_HANDLER_MAP
from django_glue.entities.template.maps import GLUE_TEMPLATE_HANDLER_MAP


CONNECTION_TO_HANDLER_MAP = {
    GlueConnection.MODEL_OBJECT: GLUE_MODEL_OBJECT_HANDLER_MAP,
    GlueConnection.QUERY_SET: GLUE_QUERY_SET_HANDLER_MAP,
    GlueConnection.FUNCTION: GLUE_FUNCTION_HANDLER_MAP,
    GlueConnection.TEMPLATE: GLUE_TEMPLATE_HANDLER_MAP
}



from django.db import models
from enum import Enum


class FieldType(str, Enum):
    AUTO_FIELD = models.AutoField.__name__
    BIG_AUTO_FIELD = models.BigAutoField.__name__
    SMALL_AUTO_FIELD = models.SmallAutoField.__name__
    BOOLEAN = models.BooleanField.__name__
    CHAR = models.CharField.__name__
    COMMA_SEPARATED_INTEGER = models.CommaSeparatedIntegerField.__name__
    DATE = models.DateField.__name__
    DATETIME = models.DateTimeField.__name__
    DECIMAL = models.DecimalField.__name__
    DURATION = models.DurationField.__name__
    EMAIL = models.EmailField.__name__
    FILE = models.FileField.__name__
    FILE_PATH = models.FilePathField.__name__
    FLOAT = models.FloatField.__name__
    FOREIGN_KEY = models.ForeignKey.__name__
    ONE_TO_ONE = models.OneToOneField.__name__
    MANY_TO_MANY = models.ManyToManyField.__name__
    INTEGER = models.IntegerField.__name__
    BIG_INTEGER = models.BigIntegerField.__name__
    SMALL_INTEGER = models.SmallIntegerField.__name__
    IP_ADDRESS = models.IPAddressField.__name__
    GENERIC_IP_ADDRESS = models.GenericIPAddressField.__name__
    JSON = models.JSONField.__name__
    NULL_BOOLEAN = models.NullBooleanField.__name__
    POSITIVE_BIG_INTEGER = models.PositiveBigIntegerField.__name__
    POSITIVE_INTEGER = models.PositiveIntegerField.__name__
    POSITIVE_SMALL_INTEGER = models.PositiveSmallIntegerField.__name__
    SLUG = models.SlugField.__name__
    TEXT = models.TextField.__name__
    TIME = models.TimeField.__name__
    URL = models.URLField.__name__
    UUID = models.UUIDField.__name__



from abc import ABC, abstractmethod
from typing import Optional

from django_glue.access.access import GlueAccess


class GlueRequestHandler(ABC):
    action: 'GlueAction' = None
    _session_data_class: 'GlueSessionData' = None
    _post_data_class: Optional['EntityBodyData'] = None

    def __init__(self, glue_session: 'GlueSession', glue_body_data: 'GlueBodyData'):
        if self._session_data_class is None:
            raise ValueError(f'Please initialize class variable _session_data_class on {self.__class__.__name__}')

        if self.action is None:
            raise ValueError(f'Please initialize class variable action on {self.__class__.__name__}')

        self.unique_name = glue_body_data.unique_name  # Unique name maps what the user is requesting

        if self._post_data_class is None:
            self.post_data = glue_body_data.data
        else:
            self.post_data = self._post_data_class(**glue_body_data.data['data'])  # The data we are expecting in post

        self.session_data = self._session_data_class(**glue_session[self.unique_name])  # data we stored in glue session.

    def has_access(self):
        glue_access = GlueAccess(self.session_data.access)
        return glue_access.has_access(self.action.required_access())

    @abstractmethod
    def process_response_data(self) -> 'GlueJsonResponseData':
        # Todo: Do we want to handle an error message here or let the system crash?
        pass



from __future__ import annotations

from enum import Enum


class GlueAccess(str, Enum):
    # The order of these variables controls how the permission cascade each other in the has_access method
    VIEW = 'view'
    CHANGE = 'change'
    DELETE = 'delete'

    def __str__(self):
        return self.value

    def has_access(self, access_required: GlueAccess):
        glue_access_tuple = tuple(GlueAccess.__members__.values())
        if glue_access_tuple.index(self) >= glue_access_tuple.index(access_required):
            return True
        else:
            return False




from enum import Enum


class GlueConnection(str, Enum):
    MODEL_OBJECT = 'model_object'
    QUERY_SET = 'query_set'
    TEMPLATE = 'template'
    FUNCTION = 'function'

    def __str__(self):
        return self.value





from django.http import HttpResponse
from django.template.loader import render_to_string

from django_glue.access.access import GlueAccess
from django_glue.entities.base_entity import GlueEntity
from django_glue.entities.template.response_data import GlueTemplateJsonData
from django_glue.entities.template.session_data import TemplateSessionData
from django_glue.handler.enums import GlueConnection


class GlueTemplate(GlueEntity):
    def __init__(
            self,
            unique_name: str,
            template_name: str,
    ):
        super().__init__(unique_name, GlueConnection.TEMPLATE, GlueAccess.VIEW)
        self.template_name = template_name

    def render_to_string(self, context_data: dict) -> str:
        return render_to_string(self.template_name, context_data)

    def to_session_data(self) -> TemplateSessionData:
        return TemplateSessionData(
            unique_name=self.unique_name,
            connection=self.connection,
            access=self.access,
            template_name=self.template_name
        )

    def to_response_data(self, rendered_template: str) -> GlueTemplateJsonData:
        return GlueTemplateJsonData(rendered_template)



from django_glue.access.access import GlueAccess
from django_glue.access.actions import GlueAction


class GlueModelObjectAction(GlueAction):
    GET = 'get'
    UPDATE = 'update'
    DELETE = 'delete'
    METHOD = 'method'

    def required_access(self) -> GlueAccess:
        if self.value in ['get', 'method']:
            return GlueAccess.VIEW
        elif self.value in ['update']:
            return GlueAccess.CHANGE
        elif self.value == 'delete':
            return GlueAccess.DELETE
        else:
            raise ValueError('That is not a valid action on a glue model object.')



import json
from dataclasses import dataclass
from tests import forms


def get_complex_form_processor(form_data):
    if form_data.get('location') == 'NYC':
        return forms.NewYorkForm(data=form_data)

    elif form_data.get('location') == 'CHI':
        return forms.ChicagoForm(data=form_data)




import json
import logging

from django.http import JsonResponse
from django.views.decorators.http import require_http_methods

from django_glue.core.decorators import require_content_types
from django_glue.handler.body_data import GlueBodyData
from django_glue.handler.utils import process_glue_request
from django_glue.response.responses import generate_json_404_response
from django_glue.session import GlueKeepLiveSession, GlueSession


@require_http_methods(["POST"])
@require_content_types('application/json', 'text/html')
def glue_data_ajax_handler_view(request):
    glue_session = GlueSession(request)
    glue_body_data = GlueBodyData(request.body)

    if glue_body_data.unique_name in glue_session.session:
        logging.warning(request.body.decode('utf-8'))
        return process_glue_request(glue_session, glue_body_data).to_django_json_response()
    else:
        return generate_json_404_response()


def glue_keep_live_handler_view(request):
    data = json.loads(request.body)
    unique_names = data.get('unique_names', [])

    if len(unique_names) > 0:
        GlueKeepLiveSession(request).update_unique_names(unique_names)

    return JsonResponse(
        data=GlueSession(request).session
    )


def glue_session_data_view(request):
    glue_session = GlueSession(request)
    return JsonResponse(
        data=glue_session.session
    )





from typing import Union, Any

from django.db.models import Model

from django_glue.access.access import GlueAccess
from django_glue.entities.base_entity import GlueEntity
from django_glue.entities.model_object.fields.entities import GlueModelFields
from django_glue.entities.model_object.fields.factories import model_object_fields_from_model
from django_glue.entities.model_object.fields.utils import get_field_value_from_model_object
from django_glue.entities.model_object.session_data import GlueModelObjectSessionData
from django_glue.handler.enums import GlueConnection
from django_glue.utils import check_valid_method_kwargs, type_set_method_kwargs


class GlueModelObject(GlueEntity):
    def __init__(
            self,
            unique_name: str,
            model_object: Model,
            access: Union[GlueAccess, str] = GlueAccess.VIEW,
            included_fields: tuple = ('__all__',),
            excluded_fields: tuple = ('__none__',),
            included_methods: tuple = ('__none__',),
    ):
        super().__init__(unique_name, GlueConnection.MODEL_OBJECT, access)

        self.model_object = model_object
        self.model = model_object._meta.model

        self.included_fields = included_fields
        self.excluded_fields = excluded_fields
        self.included_methods = included_methods

        self.fields: GlueModelFields = self.generate_field_data()

    def call_method(self, method_name, method_kwargs):
        if method_name in self.included_methods and hasattr(self.model, method_name):
            method = getattr(self.model_object, method_name)

            if check_valid_method_kwargs(method, method_kwargs):
                type_set_kwargs = type_set_method_kwargs(method, method_kwargs)

                return method(**type_set_kwargs)

        return None

    def generate_field_data(self, include_values: bool = True) -> GlueModelFields:

        glue_model_fields = model_object_fields_from_model(
            model=self.model,
            included_fields=self.included_fields,
            excluded_fields=self.excluded_fields
        )

        if include_values:
            for field in glue_model_fields:
                field.value = get_field_value_from_model_object(self.model_object, field)

        return glue_model_fields

    def generate_method_data(self):
        methods_list = list()

        for method in self.included_methods:
            if hasattr(self.model_object, method):
                methods_list.append(method)
            elif method == '__none__':
                pass
            else:
                raise KeyError(f'Method "{method}" is invalid for model type "{self.model.__class__.__name__}"')

        return methods_list

    def to_session_data(self) -> GlueModelObjectSessionData:
        return GlueModelObjectSessionData(
            connection=self.connection,
            access=self.access,
            unique_name=self.unique_name,
            fields=self.generate_field_data(include_values=False).to_dict(),
            app_label=self.model_object._meta.app_label,
            model_name=self.model_object._meta.model_name,
            object_pk=self.model_object.pk,
            included_fields=self.included_fields,
            exclude_fields=self.excluded_fields,
            methods=self.generate_method_data(),
        )

    def update(self, updated_fields: dict[str, Any]):
        for key, value in updated_fields.items():
            if hasattr(self.model_object, key):
                setattr(self.model_object, key, value)

        self.model_object.save()

        self.fields = self.generate_field_data()



from django.apps import apps
from django.db.models import Model, QuerySet

from django_glue.entities.model_object.entities import GlueModelObject
from django_glue.entities.model_object.session_data import GlueModelObjectSessionData
from django_glue.entities.query_set.session_data import GlueQuerySetSessionData


def glue_model_object_from_glue_session(glue_session: GlueModelObjectSessionData) -> GlueModelObject:
    model = apps.get_model(glue_session.app_label, glue_session.model_name)

    try:
        model_object = model.objects.get(pk=glue_session.object_pk)
    except model.DoesNotExist:
        model_object = model()

    return GlueModelObject(
        unique_name=glue_session.unique_name,
        model_object=model_object,
        access=glue_session.access,
        included_fields=glue_session.included_fields,
        excluded_fields=glue_session.exclude_fields,
        included_methods=glue_session.methods,
    )


def glue_model_object_from_glue_query_set_session(
        model_object: Model,
        query_set_session_data: GlueQuerySetSessionData
):
    return GlueModelObject(
        unique_name=query_set_session_data.unique_name,
        access=query_set_session_data.access,
        model_object=model_object,
        included_fields=query_set_session_data.included_fields,
        excluded_fields=query_set_session_data.excluded_fields,
        included_methods=query_set_session_data.included_methods,
    )


def glue_model_objects_from_query_set(
        query_set: QuerySet,
        query_set_session_data: GlueQuerySetSessionData
) -> list[GlueModelObject]:

    return [glue_model_object_from_glue_query_set_session(
        model_object,
        query_set_session_data
    ) for model_object in query_set]



from django_glue.access.access import GlueAccess
from django_glue.access.actions import GlueAction


class GlueFunctionAction(GlueAction):
    CALL = 'call'

    def required_access(self) -> GlueAccess:
        if self.value == 'call':
            return GlueAccess.VIEW
        else:
            raise ValueError('That is not a valid action on a glue query set.')



from typing import Any

from django.utils import timezone


def serialize_field_value(glue_model_field: 'GlueModelField') -> Any:
    formatted_value = glue_model_field.value

    if formatted_value is not None:

        if glue_model_field._meta.type == 'DateTimeField':
            try:
                formatted_value = timezone.localtime(glue_model_field.value).strftime('%Y-%m-%dT%H:%M')
            except Exception:
                formatted_value = glue_model_field.value.strftime('%Y-%m-%dT%H:%M')
        elif glue_model_field._meta.type == 'DateField':
            try:
                formatted_value = timezone.localdate(glue_model_field.value).strftime('%Y-%m-%d')
            except Exception:
                formatted_value = glue_model_field.value.strftime('%Y-%m-%d')

    return formatted_value



from typing import Optional

from django_glue.form.field.attrs.entities import GlueFieldAttrs
from django_glue.form.field.entities import GlueFormField
from django_glue.form.field.attrs.builder import glue_field_attr_from_model_field


class GlueFormFieldFactory:
    def __init__(self, model_field):
        self.model_field = model_field

    def attrs(self) -> GlueFieldAttrs:
        return glue_field_attr_from_model_field(self.model_field)

    def choices(self) -> list:
        if self.model_field.choices:
            if self.model_field.blank:
                return [(False, '----------')] + self.model_field.choices
            else:
                return self.model_field.choices
        else:
            if self.model_field.get_internal_type() == 'BooleanField':
                return [(True, 'Yes'), (False, 'No')]
            else:
                return [(False, '----------')]

    def factory_method(self):
        return GlueFormField(
            name=self.model_field.name,
            type=self.model_field.get_internal_type(),
            label=self.label(),
            help_text=self.help_text(),
            choices=self.choices(),
            attrs=self.attrs()
        )

    def help_text(self) -> str:
        if self.model_field.hidden or not self.model_field.help_text:
            return ''
        else:
            return str(self.model_field.help_text) if self.model_field.help_text else None

    def label(self) -> Optional[str]:
        if self.model_field.hidden:
            return None
        else:
            return str(self.model_field.verbose_name).title()



from typing import Union

from django.db.models import Model
from django.db.models.query import QuerySet
from django.http import HttpRequest

from django_glue.entities.base_entity import GlueEntity
from django_glue.entities.function.entities import GlueFunction
from django_glue.entities.model_object.entities import GlueModelObject
from django_glue.entities.query_set.entities import GlueQuerySet
from django_glue.entities.template.entities import GlueTemplate
from django_glue.session import GlueSession, GlueKeepLiveSession
from django_glue.utils import encode_unique_name


def _glue_entity(request: HttpRequest, glue_entity: GlueEntity):
    glue_session = GlueSession(request)
    glue_session.add_glue_entity(glue_entity)

    glue_keep_live_session = GlueKeepLiveSession(request)
    glue_keep_live_session.set_unique_name(glue_entity.unique_name)

    # Todo: Check to see if this has actually changed.
    glue_session.set_modified()


def glue_function(
        request: HttpRequest,
        unique_name: str,
        target: str,
):
    glue_function_entity = GlueFunction(
        unique_name=encode_unique_name(request, unique_name),
        function_path=target
    )
    _glue_entity(request, glue_function_entity)


def glue_model(
        request: HttpRequest,
        unique_name: str,
        target: Model,
        access: str = 'view',
        fields: Union[list, tuple] = ('__all__',),
        exclude: Union[list, tuple] = ('__none__',),
        methods: Union[list, tuple] = ('__none__',),
):
    glue_model_object_entity = GlueModelObject(
        unique_name=encode_unique_name(request, unique_name),
        model_object=target,
        access=access,
        included_fields=fields,
        excluded_fields=exclude,
        included_methods=methods
    )

    _glue_entity(request, glue_model_object_entity)


def glue_query_set(
        request: HttpRequest,
        unique_name: str,
        target: QuerySet,
        access: str = 'view',
        fields: Union[list, tuple] = ('__all__',),
        exclude: Union[list, tuple] = ('__none__',),
        methods: Union[list, tuple] = ('__none__',),
):
    glue_query_set_entity = GlueQuerySet(
        unique_name=encode_unique_name(request, unique_name),
        query_set=target,
        access=access,
        included_fields=fields,
        excluded_fields=exclude,
        included_methods=methods
    )

    _glue_entity(request, glue_query_set_entity)


def glue_template(
        request: HttpRequest,
        unique_name: str,
        target: str,
):

    glue_template_entity = GlueTemplate(
        unique_name=encode_unique_name(request, unique_name),
        template_name=target
    )

    _glue_entity(request, glue_template_entity)



import os
import logging
import sys

logging.basicConfig(
    format='[%(asctime)-15s] Django Glue: "%(message)s"',
    datefmt='%d/%b/%Y %H:%M:%S'
)

DEBUG = True

ALLOWED_HOSTS = ['127.0.0.1', 'localhost']

WSGI_APPLICATION = 'tests.wsgi.application'

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(BASE_DIR)

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.admin',
    'django.contrib.staticfiles',
    'django_glue',
    'tests',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django_glue.middleware.GlueMiddleware',
]

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'test.db',
    }
}

ROOT_URLCONF = 'tests.urls'

SECRET_KEY = 'django_glue_secret_key_of_secrets'

USE_TZ = True
TIME_ZONE = 'America/Edmonton'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django_glue.context_processors.glue',
            ],
            'builtins': [
            ],
            'debug': DEBUG,
        },
    },
]

STATIC_URL = '/static/'



import base64
import pickle

from django.db.models import QuerySet

from django_glue.entities.query_set.entities import GlueQuerySet
from django_glue.entities.query_set.session_data import GlueQuerySetSessionData


def decode_query_set_from_str(query_set_string) -> QuerySet:
    query = pickle.loads(base64.b64decode(query_set_string))
    decoded_query_set = query.model.objects.all()
    decoded_query_set.query = query
    return decoded_query_set


def glue_query_set_from_session_data(session_data: GlueQuerySetSessionData):
    return GlueQuerySet(
        unique_name=session_data.unique_name,
        query_set=decode_query_set_from_str(session_data.query_set_str),
        access=session_data.access,
        included_fields=session_data.included_fields,
        excluded_fields=session_data.excluded_fields,
        included_methods=session_data.included_methods,
    )



from django_glue.conf import settings


def django_glue_context_data(request) -> dict:
    context_data = {
        'django_glue_session_data': request.session.get(
            settings.DJANGO_GLUE_SESSION_NAME,
            {}
        ),
        'django_glue_keep_live_session_data': request.session.get(
            settings.DJANGO_GLUE_KEEP_LIVE_SESSION_NAME,
            {}
        ),

    }

    return context_data



from django import forms


class NewYorkForm(forms.Form):
    ketchup = forms.BooleanField(required=False)
    mustard = forms.BooleanField(required=False)
    extra_toppings = forms.CharField(required=False)
    order_quantity = forms.IntegerField(min_value=1, max_value=10)


class ChicagoForm(forms.Form):
    cheese = forms.BooleanField()
    pepperoni = forms.BooleanField(required=False)
    extra_toppings = forms.CharField(required=False)
    order_quantity = forms.IntegerField(min_value=1, max_value=5)



from django_glue.response.responses import generate_json_404_response_data


def check_access(func):
    def wrapper(self, *args, **kwargs):
        if not self.has_access():
            return generate_json_404_response_data(
                message_title='Permission Denied',
                message_body='You do not have access to this action.'
            )
        return func(self, *args, **kwargs)
    return wrapper


import inspect
import json
import urllib.parse
from typing import Optional, Callable, Union

from django.core.serializers.json import DjangoJSONEncoder


def check_valid_method_kwargs(method: Callable, kwargs: Optional[dict]):
    for kwarg in kwargs:
        if kwarg not in inspect.signature(method).parameters.keys():
            return False
    return True


def encode_unique_name(request, unique_name):
    if 'glue_encode_path' in request.GET:
        encode_path = request.GET['glue_encode_path']
    else:
        encode_path = request.path_info

    return urllib.parse.quote(f'{unique_name}|{encode_path}', safe='')


def serialize_to_json(data: Union[dict, tuple, list, str, int, float, ]) -> str:
    return json.dumps(data, cls=DjangoJSONEncoder)


def type_set_method_kwargs(method: Callable, kwargs: Optional[dict]) -> dict:
    type_set_kwargs = {}

    # This is a dict consisting of all kwargs and there type annotations (If they have type annotations)
    annotations = inspect.getfullargspec(method).annotations

    for kwarg in kwargs:
        if kwarg in annotations:
            # Converts the kwarg to match the type specified in on the methods kwargs
            type_set_kwargs[kwarg] = inspect.getfullargspec(method).annotations[kwarg](kwargs[kwarg])
        else:
            # If there is not a type annotation, the value remains the same
            type_set_kwargs[kwarg] = kwargs[kwarg]

    return type_set_kwargs



from django_glue.access.access import GlueAccess
from django_glue.access.actions import GlueAction


class GlueQuerySetAction(GlueAction):
    ALL = 'all'
    FILTER = 'filter'
    GET = 'get'
    UPDATE = 'update'
    DELETE = 'delete'
    METHOD = 'method'
    NULL_OBJECT = 'null_object'
    TO_CHOICES = 'to_choices'

    def required_access(self) -> GlueAccess:
        if self.value in ['get', 'all', 'filter', 'null_object', 'to_choices']:
            return GlueAccess.VIEW
        elif self.value in ['update', 'method']:
            return GlueAccess.CHANGE
        elif self.value == 'delete':
            return GlueAccess.DELETE
        else:
            raise ValueError('That is not a valid action on a glue query set.')



from django.db import models

from django_glue.form.field.attrs.entities import GlueFieldAttrs
from django_glue.form.field.attrs.maps import FIELD_TYPE_TO_GLUE_ATTR_FACTORY


def glue_field_attr_from_model_field(model_field) -> GlueFieldAttrs:
    try:
        glue_factory = FIELD_TYPE_TO_GLUE_ATTR_FACTORY[model_field.__class__.__name__]
    except KeyError:
        glue_factory = FIELD_TYPE_TO_GLUE_ATTR_FACTORY[models.CharField.__name__]

    return glue_factory(model_field).factory_method()



from dataclasses import dataclass, field
from typing import Union


@dataclass
class GlueFieldAttr:
    name: str
    value: Union[str, int, bool, None] = None

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'value': self.value
        }


@dataclass
class GlueFieldAttrs:
    attrs: list[GlueFieldAttr] = field(default_factory=list)

    def __add__(self, other):
        if isinstance(other, GlueFieldAttrs):
            combined_attrs = self._merge_attrs(other.attrs)
        elif isinstance(other, GlueFieldAttr):
            combined_attrs = self._merge_attrs([other])
        else:
            raise TypeError(f'Unsupported type "{type(other)}"')
        return GlueFieldAttrs(attrs=combined_attrs)

    def _merge_attrs(self, new_attrs: list[GlueFieldAttr]) -> list[GlueFieldAttr]:
        attr_dict: dict[str, GlueFieldAttr] = {attr.name: attr for attr in self.attrs}
        for attr in new_attrs:
            attr_dict[attr.name] = attr
        return list(attr_dict.values())

    def to_dict(self) -> list:
        return [attr.to_dict() for attr in self.attrs]



# Generated by Django 5.0.6 on 2024-05-24 19:57

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_bigtestmodel_foreign_key_alter_testmodel_birth_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='testmodel',
            name='email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
        migrations.AddField(
            model_name='testmodel',
            name='personality_type',
            field=models.CharField(choices=[('int', 'Introvert'), ('ext', 'Extrovert')], default='int', max_length=3),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='testmodel',
            name='favorite_number',
            field=models.IntegerField(validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(999)]),
        ),
    ]



from django import template

register = template.Library()


@register.inclusion_tag('django_glue/django_glue.html', takes_context=True)
def glue_init(context):
    return context


@register.inclusion_tag('django_glue/django_glue_bootstrap_css.html')
def glue_bootstrap_css(): ...

  
@register.inclusion_tag('django_glue/django_glue_bootstrap_js.html')
def glue_bootstrap_js(): ...


@register.inclusion_tag('django_glue/django_glue_alpine_js.html')
def glue_alpine_js(): ...


@register.simple_tag()
def to_snake_case(label: str) -> str:
  return label.replace(' ', '_').lower()



import json
import logging

from django.shortcuts import render
from django.template.response import TemplateResponse
from django.views.generic import TemplateView

from tests.models import TestModel, BigTestModel, UuidTestModel
from tests.processors import get_complex_form_processor
from tests.utils import generate_randomized_test_model, generate_big_test_model
from tests.context_data import django_glue_context_data

from django_glue.glue import glue_model, glue_query_set, glue_template, glue_function


def big_model_object_view(request):
    big_model = generate_big_test_model()
    glue_model(request, 'big_model', big_model, 'delete', fields=('foreign_key',))
    glue_query_set(request, 'big_model_query', BigTestModel.objects.all(), 'delete', fields=('foreign_key',))

    return TemplateResponse(request, template='page/big_model_object_page.html')


class ModelObjectView(TemplateView):
    template_name = 'page/model_object_page.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data.update(django_glue_context_data(self.request))

        test_model_object = generate_randomized_test_model()

        glue_model(self.request, 'test_model_1', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
        # glue_model(self.request, 'test_model_2', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))
        # glue_model(self.request, 'test_model_3', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'))
        # glue_model(self.request, 'test_model_4', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))

        big_test_model_object = generate_big_test_model()

        return context_data


class QuerySetView(TemplateView):
    template_name = 'page/query_set_page.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        context_data.update(django_glue_context_data(self.request))

        glue_query_set(self.request, 'test_query_1', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'delete', exclude=('anniversary_datetime', 'birth_date'), methods=['is_lighter_than', 'get_full_name'])
        glue_query_set(self.request, 'test_query_3', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'change', exclude=('birth_date', 'anniversary_datetime'))

        context_data['model_object_id'] = TestModel.objects.filter(id__gte=1).filter(id__lte=10000).first().id
        return context_data


def query_set_list_view(request):
    glue_query_set(request, 'test_query_1', TestModel.objects.all(), 'delete')
    return TemplateResponse(
        request,
        template='page/query_set_list_page.html'
    )


class OtherView(TemplateView):
    template_name = 'page/other_glue_page.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        context_data.update(django_glue_context_data(self.request))

        other_test_model_object = generate_randomized_test_model()

        logging.warning(f'Added Other TestModel object.')

        glue_model(self.request, 'other_test_model_1', other_test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime',))

        logging.warning('Added model object glue for Other TestModel Object in write mode')

        return context_data


def benchmark_view(request):
    from time import time
    import random
    import string

    benchmarks = {}

    test_model_object = generate_randomized_test_model()

    def generate_random_string(length):
        return ''.join(random.choice(string.ascii_letters) for _ in range(length))

    start = time()
    stop = time()

    benchmarks['No Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(1)]

    start = time()
    for unique_name in random_strings:
        glue_model(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['1 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(10)]

    start = time()
    for unique_name in random_strings:
        glue_model(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['10 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(100)]

    start = time()
    for unique_name in random_strings:
        glue_model(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['100 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(1000)]

    start = time()
    for unique_name in random_strings:
        glue_model(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['1000 Glue'] = stop - start

    return TemplateResponse(request, 'page/benchmark_page.html', {'benchmarks': benchmarks})


def view_view(request):
    return TemplateResponse(request, 'page/view_page.html')


def view_card_view(request):
    test_model_object = generate_randomized_test_model()

    glue_model(
        request=request,
        unique_name='test_model_view_card',
        target=test_model_object,
        access='delete',
        exclude=('birth_date', 'anniversary_datetime', 'best_friend'),
        methods=['is_lighter_than', 'get_full_name']
    )

    return TemplateResponse(request, 'card/view_card.html')


def template_view(request):
    glue_template(request, 'button_1', 'element/button_element.html')

    return TemplateResponse(request, 'page/template_page.html')


def function_view(request):
    glue_function(request, 'function_1', 'tests.utils.test_glue_function')
    return TemplateResponse(request, 'page/function_page.html')


def form_field_view(request):
    person = generate_randomized_test_model()
    glue_model(request, 'person', person)
    glue_query_set(request, 'people', TestModel.objects.all())
    if request.method == 'POST':
        print(request.POST)
    return TemplateResponse(request, 'page/form_fields_page.html')


def complex_form_view(request):
    LOCATION_CHOICES = [
        {'key': 'NYC', 'value': 'New York'},
        {'key': 'CHI', 'value': 'Chicago'},
    ]

    context_data = {
        'location_choices': LOCATION_CHOICES,
    }

    glue_template(request, 'new_york_element', 'complex_form/element/new_york_element.html')
    glue_template(request, 'chicago_element', 'complex_form/element/chicago_element.html')

    if request.POST:
        form = get_complex_form_processor(request.POST)
        if form.is_valid():
            print('Valid!')
        else:
            context_data['initial'] = json.dumps(request.POST)
            print(form.errors)

    return render(request, 'complex_form/page/complex_form_page.html', context_data)


def complex_model_form_view(request):
    glue_query_set(request, 'test_queryset', TestModel.objects.all(), 'delete')
    context_data = {}

    return render(request, 'complex_form/page/complex_model_form_page.html', context_data)


def uuid_model_view(request):
    uuid_title_model = UuidTestModel.objects.create(title='New Title of Function')
    glue_model(request, 'uuid_title_model', uuid_title_model)
    return TemplateResponse(request, 'page/uuid_model_page.html')



from abc import ABC, abstractmethod
from typing import Union

from django_glue.access.access import GlueAccess
from django_glue.handler.enums import GlueConnection


class GlueEntity(ABC):

    def __init__(
            self,
            unique_name: str,
            connection: GlueConnection,
            access: Union[GlueAccess, str] = GlueAccess.VIEW,
    ):

        self.unique_name = unique_name
        self.connection = connection

        if isinstance(access, str):
            self.access = GlueAccess(access)
        else:
            self.access = access

    @abstractmethod
    def to_session_data(self) -> 'GlueSessionData':
        pass



from django.http import JsonResponse
from django.http import HttpResponse


def require_content_types(*content_types):
    def decorator(view_func):
        def _wrapped_view(request, *args, **kwargs):
            if request.content_type not in content_types:
                return JsonResponse({'error': 'Unsupported Media Type'}, status=415)
            return view_func(request, *args, **kwargs)
        return _wrapped_view
    return decorator


def require_query_method(view_func):
    def _wrapped_view(request, *args, **kwargs):
        if request.method != 'QUERY':
            return HttpResponse('Method Not Allowed', status=405)
        return view_func(request, *args, **kwargs)
    return _wrapped_view



# Generated by Django 4.2.7 on 2024-06-15 04:31

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0005_testmodel_bed_time'),
    ]

    operations = [
        migrations.CreateModel(
            name='UuidTestModel',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=32)),
            ],
        ),
        migrations.AlterField(
            model_name='testmodel',
            name='bed_time',
            field=models.TimeField(default='20:00'),
        ),
    ]
